import json

from pandas import DataFrame, Series

from connect.evidence import Evidence


class DataProfilerEvidence(Evidence):
    """
    Evidence for the result of pandas profiler"""

    def __init__(self, data: dict, additional_labels: dict = None, **metadata):
        self._data = data
        super().__init__("dataset_profiler", additional_labels, **metadata)

    @property
    def data(self):
        return self._data

    @property
    def base_label(self):
        return {}


class ModelProfilerEvidence(Evidence):
    """
    Evidence for the Model Profiler.

    Splits the data into info, parameters and feature names.

    Labeling contains info on where the entries come from: user vs autogenerated
    """

    def __init__(self, data: DataFrame, additional_labels: dict = None, **metadata):
        super().__init__("model_profiler", additional_labels, **metadata)
        self._data = data["results"]

    @property
    def data(self):
        parameters = self._data.loc["parameters"]
        remaining_info = self._data[
            ~self._data.index.isin(["parameters", "feature_names", "model_name"])
        ]
        output = {
            "info": remaining_info.to_dict(),
            "parameters": parameters,
        }
        if "feature_names" in self._data.index:
            output["features_names"] = self._data.loc["feature_names"]

        return output

    @property
    def base_label(self):
        return {"model_name": self._data.loc["model_name"]}



from collections import deque
from io import StringIO
from logging import FileHandler, Formatter, Handler, StreamHandler, getLogger
from os.path import join
from sys import stdout


class TailLogHandler(Handler):
    def __init__(self, log_queue):
        Handler.__init__(self)
        self.log_queue = log_queue

    def emit(self, record):
        self.log_queue.append(self.format(record))


class TailLogger(object):
    def __init__(self, maxlen):
        self._log_queue = deque(maxlen=maxlen)
        self._log_handler = TailLogHandler(self._log_queue)

    def contents(self):
        return "\n".join(self._log_queue)

    @property
    def log_handler(self):
        return self._log_handler


class Logger:
    def __init__(
        self, name, path=None, record_stream=True, logging_level="info", formatter=None
    ):
        self.file_path = None
        self.stream = None
        if formatter is None:
            formatter = Formatter(
                "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
            )
        self.formatter = formatter
        self.logger = self.setup_logger(name, logging_level)
        self.log_capture_string = StringIO()
        if record_stream:
            self.stream = self.setup_stream()
        if path:
            self.file_path = join(path, f"{name}.log")
            self.setup_file()

    def setup_logger(self, name, logging_level):
        logger = getLogger(name)
        handler = StreamHandler(stdout)
        handler.setFormatter(self.formatter)
        logger.addHandler(handler)
        logger.setLevel(logging_level)
        return logger

    def setup_stream(self, tail_length=100):
        tail = TailLogger(tail_length)
        log_handler = tail.log_handler
        log_handler.setFormatter(self.formatter)
        self.logger.addHandler(log_handler)
        return tail

    def setup_file(self):
        file_handler = FileHandler(self.file_path)
        file_handler.setFormatter(self.formatter)
        self.logger.addHandler(file_handler)


def setup_logger(name="connect", path=None, record_stream=False, logging_level="INFO"):
    tmp = Logger(name, path, record_stream, logging_level)
    return tmp.logger, tmp.stream


global_logger, global_tail = setup_logger(path=".")



"""
Credo Governance
"""

import json
import sys
import time
from pprint import pprint
from typing import List, Optional, Union

from json_api_doc import deserialize, serialize

from connect.evidence import Evidence, EvidenceRequirement
from connect.utils import (
    check_subset,
    get_version,
    global_logger,
    json_dumps,
    wrap_list,
)

from .credo_api import CredoApi
from .credo_api_client import CredoApiClient


class Governance:
    """Class to store governance data.

    Governance is used to interact with the CredoAI Governance(Report) App.
    It has two main jobs.
    1. Get evidence_requirements of use_case and policy pack.
    2. Upload evidences gathered with evidence_requirements

    Parameters
    ----------
    credo_api_client: CredoApiClient, optional
        Credo API client. Uses default Credo API client if it is None
        Default Credo API client uses `~/.credo_config` to read API server configuration.
        Please prepare `~/.credo_config` file by downloading it from CredoAI Governance App.(My Settings > Tokens)

    Examples
    --------
    If you want to use your own configuration:

        from connect.governance.credo_api_client import CredoApiClient, CredoApiConfig
        from connect.governance.goverance import Governance

        config = CredoApiConfig(
            api_key="API_KEY", tenant="credo", api_server="https://api.credo.ai"
        )
        # or using credo_config file
        config = CredoApiConfig()
        config.load_config("CREDO_CONFIG_FILE_PATH")

        client = CredoApiClient(config=config)
        governace = Governance(credo_api_client=client)

    """

    def __init__(
        self, config_path: str = None, credo_api_client: CredoApiClient = None
    ):
        """Governance object to connect Lens with Credo AI Platform

        Parameters
        ----------
        config_path : str, optional
            path to .credoconfig file. If None, points to ~/.credoconfig, by default None
        credo_api_client : CredoApiClient, optional
            If provided, overrides the API configuration defined by
            the config path, by default None
        """
        self._use_case_id: Optional[str] = None
        self._policy_pack_id: Optional[str] = None
        self._evidence_requirements: List[EvidenceRequirement] = []
        self._evidences: List[Evidence] = []
        self._model = None
        self._plan: Optional[dict] = None
        self._unique_tags: List[dict] = []

        if credo_api_client:
            client = credo_api_client
        else:
            client = CredoApiClient(config_path=config_path)

        self._api = CredoApi(client=client)

    @property
    def model(self):
        return self._model

    @property
    def requirements_satisified(self):
        return self._match_requirements()

    @property
    def registered(self):
        return bool(self._plan)

    def add_evidence(self, evidences: Union[Evidence, List[Evidence]]):
        """
        Add evidences
        """
        self._evidences += wrap_list(evidences)

    def clear_evidence(self):
        self.set_evidence([])

    def export(self, filename=None):
        """
        Upload evidences to CredoAI Governance(Report) App

        Returns
        -------
        True
            When uploading is successful with all evidence
        False
            When it is not registered yet, or evidence is insufficient
        """
        if not self._validate_export():
            return False
        to_return = self._match_requirements()

        if filename is None:
            self._api_export()
        else:
            self._file_export(filename)

        if to_return:
            export_status = "All requirements were matched."
        else:
            export_status = "Partial match of requirements."

        global_logger.info(export_status)
        return to_return

    def get_evidence(self, verbose=False):
        """
        Returns evidence that has been send to the governance object

        Parameters
        ----------
        verbose : bool, False
            if True, print out human-readable evidence requirements
        """
        if verbose:
            self._print_evidence(self._evidences)
        return self._evidences

    def get_evidence_requirements(self, tags: dict = None, verbose=False):
        """
        Returns evidence requirements. Each evidence requirement can have optional tags
        (a dictionary). Returns requirements whose tags are a subset of the model's tags.

        For example, imagine a model has two tags: {'risk': 'high', 'model_type': 'binary'}
        and three requirements with the following tags:
            req1 = {}
            req2 = {'risk': 'high'}
            req3 = {'risk': 'high', 'model_type': 'binary'}
        In this case all requirements will apply. If the model risk was was 'low' only req_1 would
        apply. If the 'model_type' was not 'binary' then only req1 and req2 would apply.

        Parameters
        ----------
        tags : dict, optional
            Tags to subset evidence requirements. If a model has been set, will default
            to the model's tags. Evidence requirements will be returned that have no
            tags or have the same tag as provided.
        verbose : bool, False
            if True, print out human-readable evidence requirements

        Returns
        -------
        List[EvidenceRequirement]
        """
        if tags is None:
            tags = self.get_model_tags()

        reqs = [e for e in self._evidence_requirements if check_subset(e.tags, tags)]
        if verbose:
            self._print_evidence(reqs)
        return reqs

    def get_requirement_tags(self):
        """Return the unique tags used for all evidence requirements"""
        return self._unique_tags

    def get_model_tags(self):
        """Get the tags for the associated model"""
        if self._model:
            return self._model["tags"]
        else:
            return {}

    def register(
        self,
        assessment_plan_url: str = None,
        use_case_name: str = None,
        policy_pack_key: str = None,
        assessment_plan: str = None,
        assessment_plan_file: str = None,
    ):
        """
        Parameters
        ----------
        assessment_plan_url : str
            assessment plan URL
        use_case_name : str
            use case name
        policy_pack_key : str
            policy pack key
        assessment_plan : str
            assessment plan JSON string
        assessment_plan_file : str
            assessment plan file name that holds assessment plan JSON string

        Examples
        --------
        Get assessment plan and register it.
        There are three ways to do it:

            1. With assessment_plan_url.

                gov.register(assessment_plan_url="https://api.credo.ai/api/v2/tenant/use_cases/{id}/assessment_plans/{pp_id}")

            2. With use case name and policy pack key.

                gov.register(use_case_name="Fraud Detection", policy_pack_key="FAIR")

            3. With assessment_plan json string or filename. It is used in the air-gap condition.

                gov.register(assessment_plan="JSON_STRING")
                gov.register(assessment_plan_file="FILENAME")

        After successful registration, `gov.registered` returns True and able to get evidence_requirements:

            gov.registered    # returns True
            gov.get_evidence_requirements()


        """
        self._plan = None

        plan = None
        if use_case_name:
            assessment_plan_url = self._api.get_assessment_plan_url(
                use_case_name, policy_pack_key
            )

        if assessment_plan_url:
            plan = self._api.get_assessment_plan(assessment_plan_url)

        if assessment_plan:
            plan = self.__parse_json_api(assessment_plan)

        if assessment_plan_file:
            with open(assessment_plan_file, "r") as f:
                json_str = f.read()
                plan = self.__parse_json_api(json_str)

        if plan:
            self._plan = plan
            self._use_case_id = plan.get("use_case_id")
            self._policy_pack_id = plan.get("policy_pack_id")
            self._evidence_requirements = list(
                map(
                    lambda d: EvidenceRequirement(d),
                    plan.get("evidence_requirements", []),
                )
            )

            # Extract unique tags
            for x in self._evidence_requirements:
                if x.tags not in self._unique_tags:
                    self._unique_tags.append(x.tags)
            self._unique_tags = [x for x in self._unique_tags if x]

            global_logger.info(
                f"Successfully registered with {len(self._evidence_requirements)} evidence requirements"
            )

            if self._unique_tags:
                global_logger.info(
                    f"The following tags have being found in the evidence requirements: {self._unique_tags}"
                )

            self.clear_evidence()

    def set_artifacts(
        self,
        model: str,
        model_tags: dict,
        training_dataset: str = None,
        assessment_dataset: str = None,
    ):
        """Sets up internal knowledge of model and datasets to send to Credo AI Platform


        Parameters
        ----------
        model : str
            model name
        model_tags : dict
            List of key:value pairs specifying model tags. These are typically
            used to pair the model with tagged governance requirements,
            which are defined in a Governance instance's assessment_plan
        training_dataset : str, optional
            training dataset name, by default None
        assessment_dataset : str, optional
            assessment dataset name, by default None
        """

        global_logger.info(
            f"Adding model ({model}) to governance. Model has tags: {model_tags}"
        )
        prepared_model = {"name": model, "tags": model_tags}
        if training_dataset:
            prepared_model["training_dataset_name"] = training_dataset
        if assessment_dataset:
            prepared_model["assessment_dataset_name"] = assessment_dataset
        self._model = prepared_model

    def set_evidence(self, evidences: List[Evidence]):
        """
        Update evidences
        """
        self._evidences = evidences

    def tag_model(self, model):
        """Interactive utility to tag a model tags from assessment plan"""
        tags = self.get_requirement_tags()
        print(f"Select tag from assessment plan to associated with model:")
        print("0: No tags")
        for number, tag in enumerate(tags):
            print(f"{number+1}: {tag}")
        selection = int(input("Number of tag to associate: "))
        if selection == 0:
            selected_tag = None
        else:
            selected_tag = tags[selection - 1]
        print(f"Selected tag = {selected_tag}. Applying to model...")
        model.tags = selected_tag
        if self._model:
            self._model["tags"] = selected_tag

    def _api_export(self):
        global_logger.info(
            f"Uploading {len(self._evidences)} evidences.. for use_case_id={self._use_case_id} policy_pack_id={self._policy_pack_id}"
        )

        assessment = self._api.create_assessment(
            self._use_case_id, self._prepare_export_data()
        )

        if assessment:
            # wait until uploading is finished
            progress = 1
            while assessment["result"] == "in_progress":
                time.sleep(1)
                sys.stdout.write("\r")
                sys.stdout.write("." * progress)
                sys.stdout.flush()
                progress = progress + 1
                assessment = self._api.get_assessment(
                    self._use_case_id, assessment["id"]
                )

            sys.stdout.write("\r")

            # print the result
            self._assessment = assessment
            if assessment["result"] == "success":
                evidences = assessment.get("details", {}).get("evidences", [])
                duration = assessment.get("duration", 0) / 1000
                global_logger.info(
                    f"{len(evidences)} evidences were successfuly uploaded, took {duration} ms"
                )
            else:
                error = assessment["error"]
                global_logger.error(f"Error in uploading evidences : {error}")

    def _check_inclusion(self, label, evidence):
        matching_evidence = []
        for e in evidence:
            if check_subset(label, e.label):
                matching_evidence.append(e)
        if not matching_evidence:
            global_logger.info(f"Missing required evidence with label ({label}).")
            return False
        if len(matching_evidence) > 1:
            stringified_evidence = [str(e.label) for e in matching_evidence]
            nl = "\n\t\t"
            global_logger.error(
                "Multiple evidence labels were found matching one requirement.\n"
                + f"\tRequirement: {label}\n"
                + f"\tEvidences: {nl}{nl.join(stringified_evidence)}"
            )
            return False
        return matching_evidence

    def _file_export(self, filename):
        global_logger.info(
            f"Saving {len(self._evidences)} evidences to {filename}.. for use_case_id={self._use_case_id} policy_pack_id={self._policy_pack_id} "
        )
        data = self._prepare_export_data()
        meta = {"client": "Credo AI Connect", "version": get_version()}
        data = json_dumps(serialize(data=data, meta=meta))
        with open(filename, "w") as f:
            f.write(data)

    def _match_requirements(self):
        missing = []
        required_labels = [e.label for e in self.get_evidence_requirements()]
        for label in required_labels:
            matching_evidence = self._check_inclusion(label, self._evidences)
            if not matching_evidence:
                missing.append(label)
            else:
                matching_evidence[0].label = label
        return not bool(missing)

    def __parse_json_api(self, json_str):
        return deserialize(json.loads(json_str))

    def _prepare_export_data(self):
        evidences = self._prepare_evidences()
        data = {
            "policy_pack_id": self._policy_pack_id,
            "models": [self._model] if self._model else None,
            "evidences": evidences,
            "$type": "assessments",
        }
        return data

    def _prepare_evidences(self):
        evidences = list(map(lambda e: e.struct(), self._evidences))
        return evidences

    def _print_evidence(self, evidence):
        for i, label in enumerate([e.label for e in evidence]):
            print(f"\nEvidence Requirement {i}:")
            pprint(label)

    def _validate_export(self):
        if not self.registered:
            global_logger.info("Governance is not registered, please register first")
            return False

        if 0 == len(self._evidences):
            global_logger.info(
                "No evidences added to governance, please add evidences first"
            )
            return False
        return True



from copy import deepcopy

import numpy as np
import pandas as pd


class Scrubber:
    @staticmethod
    def remove_NaNs(obj):
        if isinstance(obj, pd.DataFrame):
            return Scrubber._df_remove_NaNs(obj)
        elif isinstance(obj, np.ndarray):
            return Scrubber._array_remove_NaNs(obj)
        elif isinstance(obj, dict):
            return Scrubber._dict_remove_NaNs(obj)
        elif isinstance(obj, list):
            return Scrubber._list_remove_NaNs(obj)

    @staticmethod
    def _df_remove_NaNs(data: pd.DataFrame):
        # Assume DataFrame is well-formed: does not contain lists, DFs, or other complex objects
        # returns copy of object --> no deep copy concern
        name = getattr(data, "name", None)
        scrubbed = data.fillna(np.nan).replace([np.nan], [None])
        if name:
            scrubbed.name = name
        return scrubbed

    @staticmethod
    def _list_remove_NaNs(data: list):
        return_list = deepcopy(data)
        for idx, item in enumerate(return_list):
            if isinstance(item, pd.DataFrame):
                return_list[idx] = Scrubber._df_remove_NaNs(item)
            elif isinstance(item, np.ndarray):
                return_list[idx] = Scrubber._array_remove_NaNs(item)
            elif isinstance(item, dict):
                return_list[idx] = Scrubber._dict_remove_NaNs(item)
            elif isinstance(item, list):
                return_list[idx] = Scrubber._list_remove_NaNs(item)
            else:
                # Assume no other iterable data types could be stored in list
                if item is np.nan:
                    return_list[idx] = None
        return return_list

    @staticmethod
    def _array_remove_NaNs(data: np.ndarray):
        # Assume array is well-formed: does not contain lists or other complex objects
        # returns copy of object --> no deep copy concern
        return np.where(np.isnan(data), None, data)

    @staticmethod
    def _dict_remove_NaNs(data: dict):
        return_dict = deepcopy(data)
        for key, val in return_dict.items():
            if isinstance(val, pd.DataFrame):
                return_dict[key] = Scrubber._df_remove_NaNs(val)
            elif isinstance(val, np.ndarray):
                return_dict[key] = Scrubber._array_remove_NaNs(val)
            elif isinstance(val, dict):
                return_dict[key] = Scrubber._dict_remove_NaNs(val)
            elif isinstance(val, list):
                return_dict[key] = Scrubber._list_remove_NaNs(val)
            else:
                # Assume no other iterable data types could be stored in dictionary
                if val is np.nan:
                    return_dict[key] = None
        return return_dict



import pandas as pd

from connect.evidence import EvidenceContainer, TableEvidence
from connect.utils import ValidationError

from .evidence import DeepchecksEvidence
from .utils import get_deepchecks_type

DeepChecksType = get_deepchecks_type()


class DeepchecksContainer(EvidenceContainer):
    """Container for all Table type evidence"""

    def __init__(
        self,
        name: str,
        data: DeepChecksType,
        labels: dict = None,
        metadata: dict = None,
    ):
        super().__init__(DeepchecksEvidence, data, labels, metadata)
        self.name = name

    @property
    def scrubbed_data(self):
        return self._data

    def to_evidence(self, **metadata):
        checks_2_df = {"Check_Name": list(), "Status": list()}
        for check in self.scrubbed_data.get_not_passed_checks():
            checks_2_df["Check_Name"].append(check.header)
            checks_2_df["Status"].append("Not Passed")
        for check in self.scrubbed_data.get_passed_checks():
            checks_2_df["Check_Name"].append(check.header)
            checks_2_df["Status"].append("Passed")
        for check in self.scrubbed_data.get_not_ran_checks():
            checks_2_df["Check_Name"].append(check.header)
            checks_2_df["Status"].append("Not Run")

        check_results_df = pd.DataFrame.from_dict(checks_2_df, orient="columns")
        check_results_df.name = "Lens_Deepchecks_SuiteResult"
        return [
            TableEvidence(
                self.name, check_results_df, self.labels, **self.metadata, **metadata
            )
        ]

    def _validate_inputs(self, data):
        if not isinstance(data, DeepChecksType):
            raise ValidationError("'data' must be a deepchecks.core.SuiteResult object")

    def _validate(self, data):
        pass



import requests

import connect
from connect.utils import global_logger


def get_version():
    return connect.__version__


def validate_version():
    current_version = get_version()

    package = "credoai-connect"
    response = requests.get(f"https://pypi.org/pypi/{package}/json")
    latest_version = response.json()["info"]["version"]

    on_latest = current_version == latest_version

    if not on_latest:
        global_logger.warning(
            """
            You are using credoai-connect version %s, however a newer version is available.
            Lens is updated regularly with major improvements and bug fixes.
            Please upgrade via the command: "python -m pip install --upgrade credoai-connect"
            """,
            current_version,
        )



"""
Wrappers formatting results of evaluator runs for the Credo AI Platform
"""
import pprint
from abc import ABC, abstractproperty
from datetime import datetime
from typing import Tuple

from pandas import DataFrame

from connect.utils import ValidationError


class Evidence(ABC):
    """Abstract class defining Evidence"""

    def __init__(self, type: str, additional_labels: dict = None, **metadata):
        self.type = type
        self.additional_labels = additional_labels or {}
        self.metadata = metadata
        self.creation_time: str = datetime.utcnow().isoformat()
        self._label = None
        self._validate()

    def __str__(self):
        return pprint.pformat(self.struct())

    def struct(self):
        """Structure of evidence"""
        structure = {
            "type": self.type,
            "label": self.label,
            "data": self.data,
            "generated_at": self.creation_time,
            "metadata": self.metadata,
        }
        return structure

    @property
    def label(self):
        """
        Adds evidence type specific label
        """
        # additional_labels prioritized
        if self._label is None:
            self._label = {**self.base_label, **self.additional_labels}
        return self._label

    @label.setter
    def label(self, value):
        self._label = value

    @abstractproperty
    def base_label(self):
        pass

    @abstractproperty
    def data(self):
        """
        Adds data reflecting additional structure of child classes
        """
        return {}

    def _validate(self):
        pass


class MetricEvidence(Evidence):
    """
    Evidence for Metric:value result type

    Parameters
    ----------
    type : string
        short identifier for metric.
    value : float
        metric value
    confidence_interval : [float, float]
        [lower, upper] confidence interval
    confidence_level : int
        Level of confidence for the confidence interval (e.g., 95%)
    metadata : dict, optional
        Arbitrary keyword arguments to append to metric as metadata. These will be
        displayed in the governance app
    """

    def __init__(
        self,
        type: str,
        value: float,
        confidence_interval: Tuple[float, float] = None,
        confidence_level: int = None,
        additional_labels=None,
        **metadata
    ):
        self.metric_type = type
        self.value = value
        self.confidence_interval = confidence_interval
        self.confidence_level = confidence_level
        super().__init__("metric", additional_labels, **metadata)

    @property
    def data(self):
        return {
            "value": self.value,
            "confidence_interval": self.confidence_interval,
            "confidence_level": self.confidence_level,
        }

    @property
    def base_label(self):
        label = {"metric_type": self.metric_type}
        return label

    def _validate(self):
        if self.confidence_interval and not self.confidence_level:
            raise ValidationError


class StatisticTestEvidence(Evidence):
    """
    Evidence for Statistical Test:value result type

    Parameters
    ----------
    statistic_type : string
        Short identifier for statistical test.
    value : float
        Test calculation result
    p_value : float
        p_value associated to the calculation
    significance_threshold : float
        p_value threshold to consider test significant, e.g., 0.01
    additional_labels: dict, opional
        Extra info to be added to the label section.
    metadata : dict, optional
        Arbitrary keyword arguments to append to metric as metadata. These will be
        displayed in the governance app
    """

    def __init__(
        self,
        statistic_type: str,
        test_statistic: float,
        significance_threshold: float,
        p_value: float,
        additional_labels=None,
        **metadata
    ):
        self.statistic_type = statistic_type
        self.test_statistic = test_statistic
        self.significance_threshold = significance_threshold
        self.p_value = p_value
        self.significant = (
            True if self.p_value <= self.significance_threshold else False
        )
        super().__init__("statisticTest", additional_labels, **metadata)

    @property
    def data(self):
        return {
            "test_statistic": self.test_statistic,
            "significance_threshold": self.significance_threshold,
            "p_value": self.p_value,
            "significant": self.significant,
        }

    @property
    def base_label(self):
        label = {"statistic_type": self.statistic_type}
        return label


class TableEvidence(Evidence):
    """
    Evidence for tabular data

    Parameters
    ----------
    data : str
        a pandas DataFrame to use as evidence
    metadata : dict, optional
        Arbitrary keyword arguments to append to metric as metadata. These will be
        displayed in the governance app
    """

    def __init__(
        self, name: str, table_data: DataFrame, additional_labels=None, **metadata
    ):
        self.name = name
        self._data = table_data
        super().__init__("table", additional_labels, **metadata)

    @property
    def data(self):
        columns = [
            {"value": k, "type": self._transform_type(v)}
            for k, v in self._data.dtypes.iteritems()
        ]
        return {
            "columns": columns,
            "value": self._data.values.tolist(),
        }

    @property
    def base_label(self):
        label = {"table_name": self.name}
        return label

    def _transform_type(self, pandas_type):
        lookup = {
            "int64": "number",
            "float64": "number",
            "object": "string",
            "category": "string",
            "datetime64": "datetime",
        }
        final_type = str(pandas_type)
        return lookup.get(final_type, final_type)



#
from typing import TYPE_CHECKING, TypeVar


def get_pandas_profile_type():
    """check if deepchecks is imported. If not, define object for type checking"""
    if TYPE_CHECKING:
        try:
            from pandas_profiling import ProfileReport

            pandas_profile_type = ProfileReport
        except ImportError:
            pandas_profile_type = TypeVar("Profilereport")
    else:
        from pandas_profiling import ProfileReport

        pandas_profile_type = ProfileReport
    return pandas_profile_type



"""
Everything needed to generate inputs for the Credo AI Platform

All the logic defined here is in order to standardize communication with 
Credo AI Platform API
"""
from .containers import (
    EvidenceContainer,
    MetricContainer,
    TableContainer,
    StatisticTestContainer,
)
from .evidence import Evidence, MetricEvidence, TableEvidence, StatisticTestEvidence
from .evidence_requirement import EvidenceRequirement



from functools import partial
from typing import Optional

import pandas as pd

from connect.evidence import EvidenceContainer, MetricContainer, TableContainer
from connect.governance import Governance
from connect.utils import ValidationError, wrap_list


class Adapter:
    """Basic adapter to connect with governance

    Parameters
    ----------
    governance : Governance
        the Governance instance to use for connection
    model_name : str
        Name of model
    model_tags : dict, optional
        List of key:value pairs specifying model tags. These are typically
        used to pair the model with tagged governance requirements,
        which are defined in a Governance instance's assessment_plan
    assessment_dataset_name : dict, optional
        Name of dataset used to assess the model
    """

    def __init__(
        self,
        governance: Governance,
        model_name: str,
        model_tags: Optional[dict] = None,
        assessment_dataset_name: str = None,
    ):

        self.governance = governance
        self.governance.set_artifacts(model_name, model_tags, assessment_dataset_name)

    def metrics_to_governance(
        self,
        metrics: dict,
        source: str,
        labels: dict = None,
        metadata: dict = None,
        overwrite_governance: bool = True,
    ):
        """
        Packages metrics as evidence and sends them to governance

        Parameters
        ---------
        metrics : dict or pd.DataFrame
            Dictionary of metrics. Form: {metric_type: value, ...}
        source : str
            Label for what generated the metrics. Added to metadata
        labels : dict
            Additional key/value pairs to act as labels for the evidence
        metadata : dict
            Metadata to pass to underlying evidence
        overwrite_governance : bool
            When adding evidence to a Governance object, whether to overwrite existing
            evidence or not, default False.
        """
        self._evidence_to_governance(
            self._metrics_to_evidence,
            metrics,
            source,
            labels,
            metadata,
            overwrite_governance,
        )

    def table_to_governance(
        self,
        data: dict,
        source: str,
        labels: dict = None,
        metadata: dict = None,
        overwrite_governance: bool = True,
    ):
        """
        Packages metrics as evidence and sends them to governance

        Parameters
        ---------
        data: pd.DataFrame
            Dataframe to pass to evidence_fun. The DataFrame must have a "name" attribute
        source : str
            Label for what generated the table. Added to metadata
        labels : dict
            Additional key/value pairs to act as labels for the evidence
        metadata : dict
            Metadata to pass to underlying evidence
        overwrite_governance : bool
            When adding evidence to a Governance object, whether to overwrite existing
            evidence or not, default False.
        evidence_fun : callable
            Function to pass data, labels and metadata. The function should return a list of
            evidence. Default: self._to_evidence
        """
        self._evidence_to_governance(
            TableContainer, data, source, labels, metadata, overwrite_governance
        )

    def _evidence_to_governance(
        self,
        evidence_fun,
        data,
        source,
        labels,
        metadata,
        overwrite_governance=True,
    ):
        """
        Packages data as evidence and sends to governance

        Parameters
        ---------
        evidence_fun : callable or Container
            Function to pass data, labels and metadata. The function should return a list of
            evidence. If a Container, use self._to_evidence with the specified container
        data
            data to pass to evidence_fun
        source : str
            Label for what generated the evidence. Added to metadata
        labels : dict
            Additional key/value pairs to act as labels for the evidence
        metadata : dict
            Metadata to pass to underlying evidence
        overwrite_governance : bool
            When adding evidence to a Governance object, whether to overwrite existing
            evidence or not, default False.
        """
        try:
            if issubclass(evidence_fun, EvidenceContainer):
                evidence_fun = partial(self._to_evidence, container_class=evidence_fun)
        except TypeError:
            pass
        metadata = {**(metadata or {}), "source": source}
        evidence = evidence_fun(data=data, labels=labels, metadata=metadata)
        if overwrite_governance:
            self.governance.set_evidence(evidence)
        else:
            self.governance.add_evidence(evidence)

    def _get_artifact_meta(self):
        model = self.governance.model
        if model is not None:
            model = model.copy()
            model["model_name"] = model.pop("name")
            del model["tags"]
        return model or {}

    def _metrics_to_evidence(self, data, labels=None, metadata=None):
        """Converts a dictionary of metrics to evidence

        Parameters
        ----------
        data : dict or pd.DataFrame
            Dictionary of metrics. Form: {metric_type: value, ...}
        labels : dict
            Additional labels to pass to underlying evidence
        metadata : dict
            Metadata to pass to underlying evidence

        Returns
        -------
        List
            list of Evidence
        """
        if isinstance(data, dict):
            data = pd.DataFrame(data.items(), columns=["type", "value"])
        elif not isinstance(data, pd.DataFrame):
            raise ValidationError("Metrics must be a dictionary or a dataframe")
        return self._to_evidence(MetricContainer, data, labels, metadata)

    def _to_evidence(self, container_class, data, labels, metadata):
        meta = self._get_artifact_meta()
        meta.update(metadata or {})
        container = container_class(data, labels, meta)
        return wrap_list(container.to_evidence())



import pytest

from connect.adapters import Adapter
from connect.governance import Governance

METRICS = {"precision": 0.5, "recall": 0.4}
ADAPTER = Adapter(Governance(), "model")

LABELS = [
    ({}, [{"metric_type": "precision"}, {"metric_type": "recall"}]),
    (
        {"test": "test"},
        [
            {"metric_type": "precision", "test": "test"},
            {"metric_type": "recall", "test": "test"},
        ],
    ),
]


@pytest.mark.parametrize("labels,expectation", LABELS, ids=["no_label", "label"])
def test_metrics_to_evidence(labels, expectation):
    out = ADAPTER._metrics_to_evidence(METRICS, labels)
    evidence_labels = [l.label for l in out]
    assert evidence_labels == expectation



#! /usr/bin/env python
#
# Copyright (C) 2021-2021 Credo AI
import setuptools

from connect import __version__

DESCRIPTION = "Credo AI Connect: Integration point for Credo AI Governance Platform"
DISTNAME = "credoai-connect"
MAINTAINER = "Ian Eisenberg"
MAINTAINER_EMAIL = "ian@credo.ai"
URL = ""
LICENSE = ""
DOWNLOAD_URL = "https://github.com/credo-ai/credoai_connect"
VERSION = __version__
PYTHON_REQUIRES = ">=3.8"

# Fetch ReadMe
with open("README.md", "r") as fh:
    LONG_DESCRIPTION = fh.read()

# Use requirements.txt to set the install_requires
with open("requirements.txt") as f:
    INSTALL_REQUIRES = [line.strip() for line in f]

# Fetch dev requirements files (including documentation)
with open("requirements-dev.txt") as f:
    dev_requirements = [line.strip() for line in f]

with open("docs/requirements.txt") as f:
    doc_requirements = [line.strip() for line in f]

dev_requirements += doc_requirements

EXTRAS_REQUIRES = {"dev": dev_requirements}

CLASSIFIERS = [
    "Intended Audience :: Information Technology",
    "Intended Audience :: Science/Research",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "Operating System :: OS Independent",
]

PACKAGE_DATA = {}


if __name__ == "__main__":

    import sys

    from setuptools import setup

    setup(
        name=DISTNAME,
        author=MAINTAINER,
        author_email=MAINTAINER_EMAIL,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        license=LICENSE,
        url=URL,
        version=VERSION,
        download_url=DOWNLOAD_URL,
        python_requires=PYTHON_REQUIRES,
        install_requires=INSTALL_REQUIRES,
        extras_require=EXTRAS_REQUIRES,
        packages=setuptools.find_packages(),
        classifiers=CLASSIFIERS,
        include_package_data=True,
        package_data=PACKAGE_DATA,
    )



"""
Credo API functions
"""
from typing import List

from requests.exceptions import HTTPError

from connect.utils import global_logger

from .credo_api_client import CredoApiClient


class CredoApi:
    """
    CredoApi holds Credo API functions
    """

    def __init__(self, client: CredoApiClient = None):
        self._client = client

    def set_client(self, client: CredoApiClient):
        """
        Sets Credo Api Client

        Parameters
        ----------
        client : CredoApiClient
            Credo API client
        """
        self._client = client

    def get_assessment_plan_url(self, use_case_name: str, policy_pack_key: str = None):
        """
        Convert use_case_name and policy_pack_key to assessment_plan_url

        Parameters
        ----------
        use_case_name : str
            name of a use case
        policy_pack_key : str
            policy pack key, ie: FAIR
            If it is None, it gets the first resgisterd policy pack in use case

        Returns
        -------
        None
            When use_case_name does not exist or policy_pack_key is not registered to the use_case
        str
            assessment_plan_url

        Raises
        ------
        HTTPError
            When API request returns error other than 404
        """

        try:
            path = f"assessment_plan_url?use_case_name={use_case_name}"
            if policy_pack_key:
                path += f"&policy_pack_key={policy_pack_key}"
            response = self._client.get(path)
            return response["url"]
        except HTTPError as error:
            if policy_pack_key is not None:
                global_logger.info(
                    f"Use case ({use_case_name}) with policy pack ({policy_pack_key}) does not exist"
                )
            else:
                global_logger.info(
                    f"Cannot find assessment plan URL of use case {use_case_name}"
                )
            data = error.response.json()
            errors = data.get("errors", None)
            if errors:
                detail = errors[0]["detail"]
                if error:
                    global_logger.info(f"Error : {detail}")
                else:
                    raise error

                return None
            else:
                raise error

    def get_assessment_plan(self, url: str):
        """
        Get assessment plan from API server and returns it.

        Parameters
        ----------
        url : str
            assessment plan URL

        Returns
        -------
        dict
            evidence_requirements(list): list of evidence requirements
            policy_pack_id(str): policy pack id(key+version), ie: FAIR+1
            use_case_id(str): use case id

        Raises
        ------
        HTTPError
            When API request returns error
        """

        return self._client.get(url)

    def create_assessment(self, use_case_id: str, data: dict):
        """
        Upload evidences to API server. API server creates an assessment and returns it.
        API server does the uploading process in the background, and updates assessment with the progress and result.

        Parameters
        ----------
        use_case_id : str
            use case id
        data : dict
            assessment data generated by Governance

        Returns
        -------
        dict
            id(str): assessment id

        Raises
        ------
        HTTPError
            When API request returns error
        """

        path = f"use_cases/{use_case_id}/assessments"
        return self._client.post(path, data)

    def get_assessment(self, use_case_id: str, id: str):
        """
        Get assessment to know the assessment uploading progress and result.

        Parameters
        ----------
        use_case_id : str
            use case id
        id : str
            assessment id

        Returns
        -------
        dict
            id(str): assessment id
            result(str): the result of assessment uploading, one of in_progress, success, error
            logs(list): list of logs
            details(dict): rhe summary of uploads when the result is success
            error(str): error reason when the result is error
            duration(int): time taken for uploading

        Raises
        ------
        HTTPError
            When API request returns error
        """
        path = f"use_cases/{use_case_id}/assessments/{id}"
        return self._client.get(path)



"""
Credo AI Connect package
"""
from connect.adapters import Adapter
from connect.governance import Governance
from connect.utils.version_check import validate_version

__version__ = "0.0.7"

__all__ = ["governance", "evidence", "utils"]

validate_version()
